﻿-- Регистрация мода как специализации
RealisticTerrainPhysics = {}
local RealisticTerrainPhysics_mt = Class(RealisticTerrainPhysics, Mission00)

-- Включение режима отладки
local debugMode = true -- Установите true для вывода сообщений в консоль

-- Глобальные настройки поверхности
local SurfaceSettings = {
    plowed = { friction = 0.2, resistance = 1.8, deformation = 1.0, stuckChance = 0.5, slipChance = 0.8 },
    cultivated = { friction = 0.3, resistance = 1.5, deformation = 0.8, stuckChance = 0.3, slipChance = 0.6 },
    mud = { friction = 0.1, resistance = 2.0, deformation = 1.0, stuckChance = 0.4, slipChance = 0.5 },
    grass = { friction = 0.3, resistance = 1.5, deformation = 0.6, stuckChance = 0.1, slipChance = 0.2 },
    asphalt = { friction = 0.8, resistance = 1.1, deformation = 0.2, stuckChance = 0.0, slipChance = 0.0 },
    dirt_road = { friction = 0.7, resistance = 1.4, deformation = 0.4, stuckChance = 0.2, slipChance = 0.3 },
    natural_cover = { friction = 0.6, resistance = 1.3, deformation = 0.5, stuckChance = 0.2, slipChance = 0.4 },
    default = { friction = 0.8, resistance = 1.1, deformation = 0.2, stuckChance = 0.1, slipChance = 0.2 }
}

-- Временные параметры тестирования
local testingMode = false
local testSlipChance = 0.9
local testStuckChance = 0.9

-- Инициализация мода
function RealisticTerrainPhysics.initSpecialization()
    if debugMode then
        print("RealisticTerrainPhysics: Initializing specialization")
    end
end

-- Регистрация мода
function RealisticTerrainPhysics.registerSpecialization()
    SpecializationUtil.registerSpecialization("realisticTerrainPhysics", "RealisticTerrainPhysics", RealisticTerrainPhysics)
end

-- Регистрация событий
function RealisticTerrainPhysics.registerEventListeners(vehicleType)
    SpecializationUtil.registerEventListener(vehicleType, "onLoad", RealisticTerrainPhysics)
    SpecializationUtil.registerEventListener(vehicleType, "onUpdate", RealisticTerrainPhysics)
    SpecializationUtil.registerEventListener(vehicleType, "onUpdateTick", RealisticTerrainPhysics)
end

-- Загрузка мода для транспортного средства
function RealisticTerrainPhysics:onLoad(savegame)
    self.realisticTerrainPhysics = {}
    if debugMode then
        print("RealisticTerrainPhysics: Loaded for vehicle " .. tostring(self:getName()))
    end
end

-- Изменение параметров деформации почвы
function RealisticTerrainPhysics:adjustTerrainDisplacement(isRaining, vehicle)
    local weightFactor = vehicle:getMass() / 6000
    local rainIntensityFactor = isRaining and g_currentMission.environment.weather:getRainFallScale() or 0
    local surfaceType = self:getSurfaceType(vehicle) or "default"
    local deformationFactor = SurfaceSettings[surfaceType].deformation

    local maxDisplacement = 0.7 + (weightFactor * 0.9) + (rainIntensityFactor * 0.9) + deformationFactor

    if vehicle.spec_wheels then
        for _, wheel in pairs(vehicle.spec_wheels.wheels) do
            if wheel.width < 0.3 then -- Узкие шины
                maxDisplacement = maxDisplacement + 0.7
            elseif wheel.width < 0.5 then -- Стандартные шины
                maxDisplacement = maxDisplacement + 0.5
            elseif wheel.width < 0.7 then -- Широкие шины
                maxDisplacement = maxDisplacement + 0.3
            else -- Двойные шины
                maxDisplacement = maxDisplacement + 0.2
            end
        end
    end

    -- Применяем деформацию через Giants Engine
    -- Примечание: Giants Engine не предоставляет прямого метода для изменения деформации в реальном времени,
    -- поэтому используем косвенное влияние через силу сцепления
    if debugMode then
        print("Max displacement calculated to " .. maxDisplacement .. " for surface: " .. surfaceType)
    end
end

-- Изменение мягкости почвы
function RealisticTerrainPhysics:adjustFieldSoftness(isRaining, vehicle)
    local baseSoftness = isRaining and 0.1 or 0.4
    local weightFactor = vehicle:getMass() / 6000
    local resistance = baseSoftness - (weightFactor * 0.3)

    -- Применяем сопротивление через силу сцепления колес
    if vehicle.spec_wheels then
        for _, wheel in pairs(vehicle.spec_wheels.wheels) do
            wheel.groundFrictionCoefficient = resistance
        end
    end

    if debugMode then
        print("Field softness set to " .. resistance)
    end
end

-- Применение силы тяжести, инерции, регулировка разгона и динамической инерции
function RealisticTerrainPhysics:applyGravityAndInertia(vehicle, isRaining)
    if not vehicle.spec_motorized or not vehicle.spec_motorized.motor or not vehicle.spec_drivable then
        return -- Пропускаем, если нет двигателя или управления
    end

    local mass = vehicle:getMass() or 3000
    local gravity = 9.8
    local surfaceType = self:getSurfaceType(vehicle) or "default"
    local settings = SurfaceSettings[surfaceType] or SurfaceSettings["default"]

    -- Базовый коэффициент разгона
    local baseAcceleration = vehicle.spec_motorized.motor:getAccelerationLimit() or 1.0
    local maxAcceleration = baseAcceleration -- Сохраняем максимальное ускорение

    -- Текущая и максимальная скорость
    local currentSpeed = vehicle.spec_drivable.lastSpeed * 3600 / 1000 -- км/ч
    local maxSpeed = vehicle.spec_motorized.motor.maxForwardSpeed or 20 -- По умолчанию 20 км/ч
    local speedRatio = math.max(0, math.min(currentSpeed / maxSpeed, 1.0)) -- Отношение 0..1

    -- Учёт массы (чем тяжелее, тем медленнее разгон)
    local massFactor = math.max(1.0, mass / 3000) -- Нормализация: 3000 кг = стандарт
    local accelerationReduction = 0.5 / massFactor -- Тяжёлые машины теряют до 50% ускорения

    -- Учёт поверхности
    local surfaceResistance = settings.resistance
    if isRaining then
        surfaceResistance = surfaceResistance * (1.0 + 0.3 * (g_currentMission.environment.weather:getRainFallScale() or 0))
    end
    accelerationReduction = accelerationReduction * (1.0 + surfaceResistance * 0.2) -- Мягкие поверхности замедляют разгон

    -- Учёт нагрузки
    local loadWeight = vehicle.spec_fillUnit and vehicle.spec_fillUnit.fillUnits[1].fillLevel or 0
    local loadFactor = 1.0 + (loadWeight / 5000) -- 5000 кг груза = +100% замедления
    accelerationReduction = accelerationReduction * loadFactor

    -- Учёт инерции (статическая часть)
    local inertiaFactor = 0.05 * mass
    accelerationReduction = accelerationReduction + (inertiaFactor * gravity * 0.005)

    -- Динамическая инерция (зависит от текущей скорости)
    local inertiaMultiplier = 1.0 + (speedRatio * 0.5) -- При высокой скорости (+50% к замедлению)
    if speedRatio < 0.2 then
        inertiaMultiplier = 0.8 -- При низкой скорости ускорение чуть быстрее
    end
    accelerationReduction = accelerationReduction * inertiaMultiplier

    -- Учёт торможения
    local brakingFactor = 1.0
    if vehicle.spec_drivable.isBraking then -- Если игрок тормозит
        brakingFactor = 1.0 + (speedRatio * 0.7) -- Высокая скорость увеличивает время торможения
        if settings.slipChance > 0.4 or surfaceType == "mud" then
            brakingFactor = brakingFactor * 1.2 -- Скользкие поверхности замедляют торможение
        end
        accelerationReduction = accelerationReduction * brakingFactor
    end

    -- Применяем новое ускорение
    local newAcceleration = math.max(maxAcceleration * (1.0 / accelerationReduction), 0.1) -- Не ниже 0.1
    vehicle.spec_motorized.motor:setAccelerationLimit(newAcceleration)

    if debugMode then
        print(string.format("Acceleration adjusted for %s: mass=%.0f, surface=%s, load=%.0f, speedRatio=%.2f, inertiaMult=%.2f, braking=%s, oldAccel=%.2f, newAccel=%.2f",
            tostring(vehicle:getName()), mass, surfaceType, loadWeight, speedRatio, inertiaMultiplier, tostring(vehicle.spec_drivable.isBraking), maxAcceleration, newAcceleration))
    end
end

-- Учет угла наклона
function RealisticTerrainPhysics:adjustMovementOnSlope(vehicle)
    if not vehicle.spec_drivable or not vehicle.spec_wheels then
        return -- Пропускаем, если транспортное средство не управляемое или без колёс
    end

    -- Получаем позицию и направление транспортного средства
    local x, y, z = getWorldTranslation(vehicle.rootNode)
    local dirX, dirY, dirZ = localDirectionToWorld(vehicle.rootNode, 0, 0, 1) -- Направление вперёд
    local forwardX, forwardZ = x + dirX * 2, z + dirZ * 2 -- Точка впереди на 2 метра
    local rearX, rearZ = x - dirX * 2, z - dirZ * 2 -- Точка сзади на 2 метра

    -- Вычисляем высоту спереди и сзади
    local heightForward = getTerrainHeightAtWorldPos(g_currentMission.terrainRootNode, forwardX, 0, forwardZ)
    local heightRear = getTerrainHeightAtWorldPos(g_currentMission.terrainRootNode, rearX, 0, rearZ)

    -- Вычисляем угол наклона (в градусах)
    local heightDiff = heightForward - heightRear
    local slopeAngle = math.deg(math.atan(heightDiff / 4)) -- 4 = расстояние между точками (2 + 2)

    local mass = vehicle:getMass() or 5000
    local gravityForce = mass * 9.8 * math.sin(math.rad(slopeAngle))

    -- Получаем текущую скорость в км/ч
    local currentSpeed = vehicle.spec_drivable.lastSpeed * 3600 / 1000
    local newSpeed = currentSpeed

    -- Определяем максимальную скорость транспортного средства
    local maxSpeed = vehicle.spec_motorized and vehicle.spec_motorized.motor.maxForwardSpeed or 20 -- По умолчанию 20 км/ч

    -- Получаем тип поверхности и погодные условия
    local isRaining = g_currentMission.environment.weather:getIsRaining()
    local surfaceType = self:getSurfaceType(vehicle) or "default"
    local settings = SurfaceSettings[surfaceType] or SurfaceSettings["default"]
    local slipFactor = settings.slipChance -- Используем slipChance для определения скользкости

    if isRaining then
        slipFactor = slipFactor * (1.0 + 0.5 * (g_currentMission.environment.weather:getRainFallScale() or 0))
    end

    if slopeAngle > 0 then
        -- Подъём: замедляем транспортное средство
        newSpeed = currentSpeed - (gravityForce * 0.0001)
        newSpeed = math.max(newSpeed, 0) -- Не допускаем отрицательной скорости
    elseif slopeAngle < 0 then
        -- Спуск: ускоряем транспортное средство
        newSpeed = currentSpeed + (math.abs(gravityForce) * 0.0001)

        -- Добавляем влияние скользкой поверхности
        if slipFactor > 0.4 or surfaceType == "mud" then -- Скользкие поверхности или грязь
            local randomSlip = (math.random() - 0.5) * 0.2 * slipFactor -- Случайное колебание ±10% * slipFactor
            newSpeed = newSpeed * (1.0 + randomSlip)
            
            -- Снижаем трение колес для нестабильности
            if vehicle.spec_wheels then
                for _, wheel in pairs(vehicle.spec_wheels.wheels) do
                    local originalFriction = wheel.groundFrictionCoefficient or settings.friction
                    wheel.groundFrictionCoefficient = math.max(originalFriction * (1.0 - slipFactor * 0.5), 0.05)
                end
            end
        end

        newSpeed = math.min(newSpeed, maxSpeed * 1.5) -- Ограничиваем скорость до 1.5x от максимальной
    end

    -- Применяем новую скорость (переводим обратно в м/с)
    vehicle.spec_drivable.lastSpeed = newSpeed * 1000 / 3600

    if debugMode then
        print(string.format("Slope adjustment for %s: slopeAngle=%.2f, gravityForce=%.2f, surface=%s, slipFactor=%.2f, oldSpeed=%.2f, newSpeed=%.2f",
            tostring(vehicle:getName()), slopeAngle, gravityForce, surfaceType, slipFactor, currentSpeed, newSpeed))
    end
end

-- Учет загруженности
function RealisticTerrainPhysics:adjustHandlingBasedOnLoad(vehicle)
    local loadWeight = vehicle.spec_fillUnit and vehicle.spec_fillUnit.fillUnits[1].fillLevel or 0
    local baseTurningRadius = 5.0
    local adjustedTurningRadius = baseTurningRadius + loadWeight / 1000

    if vehicle.spec_drivable then
        vehicle.spec_drivable.maxTurningRadius = adjustedTurningRadius
    end

    if debugMode then
        print("Handling adjusted for vehicle " .. tostring(vehicle:getName()) .. ": turning radius = " .. adjustedTurningRadius)
    end
end

-- Изменение трения колес
function RealisticTerrainPhysics:adjustWheelFrictionBasedOnWeatherAndSurface(vehicle, isRaining)
    local surfaceType = self:getSurfaceType(vehicle) or "default"
    local settings = SurfaceSettings[surfaceType] or SurfaceSettings["default"]
    local friction = settings.friction

    if isRaining then
        local rainIntensity = g_currentMission.environment.weather:getRainFallScale() or 0
        if surfaceType == "dirt_road" then
            friction = friction * 0.7
        elseif surfaceType == "natural_cover" then
            friction = friction * 0.6
        else
            friction = friction * (1.0 - 0.4 * rainIntensity)
        end
    else
        if surfaceType == "dirt_road" then
            friction = friction * 0.9
        elseif surfaceType == "natural_cover" then
            friction = friction * 0.8
        end
    end

    if vehicle.spec_wheels then
        for _, wheel in pairs(vehicle.spec_wheels.wheels) do
            wheel.groundFrictionCoefficient = friction
        end
    end

    if debugMode then
        print("Friction adjusted for vehicle " .. tostring(vehicle:getName()) .. " on " .. surfaceType .. ": " .. friction)
    end
end

-- Симуляция пробуксовки
function RealisticTerrainPhysics:simulateCompleteSlipInsteadOfStuck(vehicle)
    if vehicle.spec_wheels then
        for _, wheel in pairs(vehicle.spec_wheels.wheels) do
            wheel.groundFrictionCoefficient = 0.1 -- Минимальное трение для пробуксовки
        end
    end
    if vehicle.spec_drivable then
        vehicle.spec_drivable.lastSpeed = 0
    end

    if debugMode then
        print("Vehicle " .. tostring(vehicle:getName()) .. " is experiencing full slip!")
    end
end

-- Проверка пробуксовки
function RealisticTerrainPhysics:checkVehicleSlipInsteadOfStuck(vehicle, isRaining)
    local surfaceType = self:getSurfaceType(vehicle) or "default"
    local settings = SurfaceSettings[surfaceType] or SurfaceSettings["default"]
    local stuckChance = testingMode and testStuckChance or settings.stuckChance

    if isRaining then
        local rainIntensity = g_currentMission.environment.weather:getRainFallScale() or 0
        stuckChance = stuckChance * (1.0 + 0.5 * rainIntensity)
    end

    if vehicle.spec_wheels and vehicle.spec_attacherJoints then
        for _, wheel in pairs(vehicle.spec_wheels.wheels) do
            if wheel.width < 0.3 and vehicle.spec_attacherJoints.attachedImplements then
                stuckChance = stuckChance + 0.2
            end
        end
    end

    if math.random() < stuckChance then
        self:simulateCompleteSlipInsteadOfStuck(vehicle)
    end
end

-- Получение типа поверхности
function RealisticTerrainPhysics:getSurfaceType(vehicle)
    -- Пример: получение типа поверхности через координаты
    local x, y, z = getWorldTranslation(vehicle.rootNode)
    local groundType = g_currentMission.fieldGroundSystem:getFieldGroundType(x, z)
    if groundType == FieldGroundType.PLOWED then
        return "plowed"
    elseif groundType == FieldGroundType.CULTIVATED then
        return "cultivated"
    elseif groundType == FieldGroundType.GRASS then
        return "grass"
    else
        return "default"
    end
end

-- Основной цикл обновления
function RealisticTerrainPhysics:onUpdate(dt)
    local isRaining = g_currentMission.environment.weather:getIsRaining()
    self:adjustWheelFrictionBasedOnWeatherAndSurface(self, isRaining)
    self:applyGravityAndInertia(self, isRaining)
    self:adjustMovementOnSlope(self)
    self:adjustHandlingBasedOnLoad(self)
    self:checkVehicleSlipInsteadOfStuck(self, isRaining)
    self:adjustTerrainDisplacement(isRaining, self)
    self:adjustFieldSoftness(isRaining, self)
end

-- Регистрация мода в игре
addModEventListener(TerrainHeightAdjuster)

-- Вывод отладочной информации
if debugMode then
    print("RealisticTerrainPhysics mod loaded successfully")
end